"""
Test the ``meta_mapper()`` function and ``MetaMapper`` class.
"""

from collections import namedtuple

import pytest

from jobbergate_api.meta_mapper import MetaField, MetaMapper


def test___init___successfully_populates_field_dict():
    """
    Provide a test case for the ``MetaMapper`` class instantiator.

    Test that a MetaMapper can be successfully instantiated.
    """
    instance = MetaMapper(
        foo=MetaField(description="foo description", example="foo example"),
        bar=MetaField(example="bar example"),
        baz=MetaField(description="baz description"),
    )
    assert instance.field_dict["foo"].description == "foo description"
    assert instance.field_dict["foo"].example == "foo example"
    assert instance.field_dict["bar"].description is None
    assert instance.field_dict["bar"].example == "bar example"
    assert instance.field_dict["baz"].description == "baz description"
    assert instance.field_dict["baz"].example is None


def test___init___fails_if_keyword_argument_is_not_a_MetaField():
    """
    Provide a test case for the ``MetaMapper`` class instantiator.

    Test that a MetaMapper cannot be instantiated if a keyword argument is not an instance of ``MetaField``.
    """
    ValidDuckField = namedtuple("ValidDuckField", ["description", "example"])
    InvalidDuckField = namedtuple("InvalidDuckField", ["description"])

    MetaMapper(foo=ValidDuckField(description="foo description", example="foo example"))  # type:ignore
    with pytest.raises(ValueError, match="Keyword argument 'foo' does not have attribute 'example'"):
        MetaMapper(foo=InvalidDuckField(description="foo description"))  # type:ignore


def test__call___remaps_fields_when_they_are_present_in_the_schema_being_mapped():
    """
    Provide a test case for calling a ``MetaMapper`` instance.

    Test that dictionary fields are remapped if they have entries in the mapper.
    """
    mapper = MetaMapper(
        foo=MetaField(
            description="new foo description",
            example="new foo example",
        ),
        bar=MetaField(
            example="new bar example",
        ),
        baz=MetaField(
            description="new baz description",
        ),
    )
    full_instance = dict(
        properties=dict(
            foo=dict(
                description="foo description",
                example="foo example",
            ),
            bar=dict(
                description="bar description",
                example="bar example",
            ),
            baz=dict(
                description="baz description",
                example="baz example",
            ),
        ),
    )
    mapper(full_instance)
    assert full_instance["properties"]["foo"]["description"] == "new foo description"
    assert full_instance["properties"]["foo"]["example"] == "new foo example"
    assert full_instance["properties"]["bar"]["description"] == "bar description"
    assert full_instance["properties"]["bar"]["example"] == "new bar example"
    assert full_instance["properties"]["baz"]["description"] == "new baz description"
    assert full_instance["properties"]["baz"]["example"] == "baz example"



"""
Provides a metadata-mapper for re-using descriptions and examples across many pydantic models.
"""

from dataclasses import dataclass, fields
from typing import Any, Dict, Optional


@dataclass
class MetaField:
    """
    Provides a dataclass that describes the metadata that will be mapped for an individual field.
    """

    description: Optional[str] = None
    example: Optional[Any] = None


class MetaMapper:
    """
    Maps re-usable metadata for fields. Should be used with the `schema_extra` property of a Model's Config.

    Example::

        foo_meta = MetaMapper(
            id=MetaField(
                description="The unique identifier of this Foo",
                example=13,
            ),
            name=MetaField(
                description="The name of this Foo",
                example="Bar",
            ),
            is_active=MetaField(
                description="Indicates if this Foo is active",
                example=True,
            ),
            created_at=MetaField(
                description="The timestamp indicating when this Foo was created",
                example="2021-12-29 11:58:00",
            ),
        )


        class CreateFooRequest(BaseModel):
            name: str
            is_active: Optional[bool]

            class Config:
                schema_extra = foo_meta


        class UpdateFooRequest(BaseModel):
            name: Optional[str] = None
            is_active: Optional[bool] = None

            class Config:
                schema_extra = foo_meta


        class FooResponse(BaseModel):
            id: int
            name: str
            is_active: bool
            created_at: DateTime

            class Config:
                schema_extra = foo_meta


    Notice in this example that the fields may be required in some models and optional in others. Further,
    not all the fields are present in all the models. The MetaMapper allows the models to share field metadata
    and yet define the fields independently.
    """

    field_dict: Dict[str, MetaField]

    def __init__(self, **kwargs: MetaField):
        """
        Map the kwargs into the field_dict.

        All kwargs *should* be MetaFields, but any object duck-typed to include all the attributes of a
        MetaField will be accepted.
        """
        required_fields = (f.name for f in fields(MetaField))
        for (key, value) in kwargs.items():
            for field in required_fields:
                if not hasattr(value, field):
                    raise ValueError(
                        f"Keyword argument '{key}' does not have attribute '{field}'. "
                        "All keyword arguments must be MetaFields (or matching duck-typed objects)"
                    )
        self.field_dict = dict(**kwargs)

    def __call__(self, schema: Dict[str, Any], *_) -> None:
        """
        Map the MetaFields onto the metadata properties of a schema.

        Should be used in a pydantic Model's Config class.
        """
        for (key, old_metadata) in schema["properties"].items():
            new_metadata = self.field_dict.get(key)
            if new_metadata is None:
                continue

            if new_metadata.description is not None:
                old_metadata["description"] = new_metadata.description

            if new_metadata.example is not None:
                old_metadata["example"] = new_metadata.example



"""
Configuration of pytest.
"""

import contextlib
import dataclasses
import datetime
import random
import string
import typing
from textwrap import dedent
from unittest.mock import patch

import pytest
import sqlalchemy
from asgi_lifespan import LifespanManager
from file_storehouse.engine import EngineLocal
from httpx import AsyncClient

from jobbergate_api.apps.applications.models import applications_table
from jobbergate_api.apps.job_scripts.models import job_scripts_table
from jobbergate_api.apps.job_submissions.models import job_submissions_table
from jobbergate_api.config import settings
from jobbergate_api.main import app
from jobbergate_api.metadata import metadata
from jobbergate_api.s3_manager import file_manager_factory
from jobbergate_api.storage import build_db_url, database

# Charset for producing random strings
CHARSET = string.ascii_letters + string.digits + string.punctuation


@pytest.fixture(autouse=True, scope="session")
async def database_engine():
    """
    Provide a fixture to get access to the database engine with a fixture.
    """
    engine = sqlalchemy.create_engine(build_db_url())
    yield engine


@pytest.fixture(autouse=True, scope="session")
async def enforce_empty_database(database_engine):
    """
    Make sure our database is empty at the end of each test.
    """
    metadata.create_all(database_engine)
    yield

    await database.connect()
    for table in (applications_table, job_scripts_table, job_submissions_table):
        count = await database.execute(sqlalchemy.select([sqlalchemy.func.count()]).select_from(table))
        assert count == 0
    await database.disconnect()

    metadata.drop_all(database_engine)


@pytest.fixture()
async def startup_event_force():
    """
    Force the async event loop to begin.
    """
    async with LifespanManager(app):
        yield


@pytest.fixture(autouse=True)
def enforce_mocked_oidc_provider(mock_openid_server):
    """
    Enforce that the OIDC provider used by armasec is the mock_openid_server provided as a fixture.

    No actual calls to an OIDC provider will be made.
    """
    yield


@pytest.fixture
async def client():
    """
    Provide a client that can issue fake requests against fastapi endpoint functions in the backend.
    """
    async with AsyncClient(app=app, base_url="http://test") as client:
        yield client


@pytest.fixture
async def inject_security_header(client, build_rs256_token):
    """
    Provide a helper method that will inject a security token into the requests for a test client.

    If no permissions are provided, the security token will still be valid but will not carry any permissions.
    Uses the `build_rs256_token()` fixture from the armasec package. If `client_id` is provided, it
    will be injected into the custom identity claims.
    """

    def _helper(
        owner_email: str,
        *permissions: typing.List[str],
        client_id: typing.Optional[str] = None,
    ):
        claim_overrides = dict(
            email=owner_email,
            client_id=client_id,
            permissions=permissions,
        )
        token = build_rs256_token(claim_overrides=claim_overrides)
        client.headers.update({"Authorization": f"Bearer {token}"})

    return _helper


@pytest.fixture
def time_frame():
    """
    Provide a fixture to use as a context manager for asserting events happened in a window of time.
    """

    @dataclasses.dataclass
    class TimeFrame:
        """
        Class for storing the beginning and end of a time frame.
        """

        now: datetime.datetime
        later: typing.Optional[datetime.datetime]

        def __contains__(self, moment: datetime.datetime):
            """
            Check if a given moment falls within a time-frame.
            """
            if self.later is None:
                return False
            return moment >= self.now and moment <= self.later

    @contextlib.contextmanager
    def _helper():
        """
        Context manager for defining the time-frame for the time_frame fixture.
        """
        window = TimeFrame(now=datetime.datetime.utcnow() - datetime.timedelta(seconds=1), later=None)
        yield window
        window.later = datetime.datetime.utcnow() + datetime.timedelta(seconds=1)

    return _helper


@pytest.fixture
def tweak_settings():
    """
    Provide a fixture to use as a context manager where the app settings may be temporarily changed.
    """

    @contextlib.contextmanager
    def _helper(**kwargs):
        """
        Context manager for tweaking app settings temporarily.
        """
        previous_values = {}
        for (key, value) in kwargs.items():
            previous_values[key] = getattr(settings, key)
            setattr(settings, key, value)
        yield
        for (key, value) in previous_values.items():
            setattr(settings, key, value)

    return _helper


@pytest.fixture
def dummy_application_source_file() -> str:
    """
    Fixture to return a dummy application source file.
    """
    return dedent(
        """
        from jobbergate_cli.application_base import JobbergateApplicationBase
        from jobbergate_cli import appform

        class JobbergateApplication(JobbergateApplicationBase):

            def mainflow(self, data):
                questions = []

                questions.append(appform.List(
                    variablename="partition",
                    message="Choose slurm partition:",
                    choices=self.application_config['partitions'],
                ))

                questions.append(appform.Text(
                    variablename="job_name",
                    message="Please enter a jobname",
                    default=self.application_config['job_name']
                ))
                return questions
        """
    ).strip()


@pytest.fixture
def dummy_template() -> str:
    """
    Fixture to return a dummy template.
    """
    return dedent(
        """
        #!/bin/bash

        #SBATCH --job-name={{data.job_name}}
        #SBATCH --partition={{data.partition}}
        #SBATCH --output=sample-%j.out


        echo $SLURM_TASKS_PER_NODE
        echo $SLURM_SUBMIT_DIR
        """
    ).strip()


@pytest.fixture
def dummy_application_config() -> str:
    """
    Fixture to return a dummy application config file.
    """
    return dedent(
        """
        application_config:
            job_name: rats
            partitions:
                - debug
                - partition1
        jobbergate_config:
            default_template: test_job_script.sh
            output_directory: .
            supporting_files:
                - test_job_script.sh
            supporting_files_output_name:
                test_job_script.sh:
                    - support_file_b.py
            template_files:
                - templates/test_job_script.sh
        """
    ).strip()


@pytest.fixture
def make_dummy_file(tmp_path):
    """
    Provide a fixture that will generate a temporary file with ``size`` random bytes of text data.
    """

    def _helper(filename, size: int = 100, content: str = ""):
        """
        Auxillary function that builds the temporary file.
        """
        if not content:
            content = "".join(random.choice(CHARSET) for _ in range(size))
        dummy_path = tmp_path / filename
        dummy_path.write_text(content)
        return dummy_path

    return _helper


@pytest.fixture
def make_files_param():
    """
    Provide a fixture to use as a context manager that builds the ``files`` parameter.

    Open the supplied file(s) and build a ``files`` param appropriate for using
    multi-part file uploads with the client.
    """

    @contextlib.contextmanager
    def _helper(*file_paths):
        """
        Context manager that opens the file(s) and yields the ``files`` param from it.
        """
        with contextlib.ExitStack() as stack:
            yield [
                (
                    "upload_files",
                    (path.name, stack.enter_context(open(path)), "text/plain"),
                )
                for path in file_paths
            ]

    return _helper


@pytest.fixture(scope="function")
def mocked_file_manager_factory(tmp_path):
    """
    Fixture to replace the default file engine (EngineS3) by a local one.

    In this way, all objects are stored in a temporary directory ``tmp_path``,
    that is yield for reference. Files and directories can than be managed
    normally with pathlib.Path.

    It also clears the cache from file_manager_factory.
    """

    def local_engine_factory(*, prefix: str, **kwargs):
        return EngineLocal(base_path=tmp_path / prefix)

    file_manager_factory.cache_clear()

    with patch("jobbergate_api.s3_manager.engine_factory", wraps=local_engine_factory):
        yield tmp_path

    file_manager_factory.cache_clear()



"""
Test schemas for Jobbergate applications.
"""

from textwrap import dedent

import pytest

from jobbergate_api.apps.applications.schemas import ApplicationConfig, JobbergateConfig


@pytest.fixture
def reference_application_config() -> str:
    """
    Fixture to return a complete application config file.
    """
    return dedent(
        """
        application_config:
            job_name: rats
            partitions:
                - debug
                - partition1
        jobbergate_config:
            default_template: test_job_script.sh
            output_directory: .
            supporting_files:
                - test_job_script.sh
                - test_another_job_script.sh
            supporting_files_output_name:
                test_job_script.sh:
                    - support_file_b.py
                test_another_job_script.sh:
                    - support_file_c.py
            template_files:
                - templates/test_job_script.sh
        """
    ).strip()


@pytest.fixture
def dummy_application_config() -> str:
    """
    Fixture to return a dummy application config file.

    Due to the business logic, this file should be parsed to an object
    to the config file above (reference_application_config).
    """
    return dedent(
        """
        application_config:
            job_name: rats
            partitions:
                - debug
                - partition1
        jobbergate_config:
            default_template: test_job_script.sh
            output_directory: .
            supporting_files_output_name:
                test_job_script.sh:
                    - support_file_b.py
                test_another_job_script.sh: support_file_c.py
            template_files:
                - templates/test_job_script.sh
        """
    ).strip()


class TestJobbergateConfig:
    """
    Test jobbergate configuration.
    """

    def test_get_supporting_files_from_supporting_files_output_name__success(self):
        """
        Test that supporting_files is obtained from the keys of supporting_files_output_name.
        """
        config = JobbergateConfig(supporting_files_output_name=dict(foo="string", bar="string"))

        assert config.supporting_files == ["foo", "bar"]

    def test_get_supporting_files_from_supporting_files_output_name__no_replacement(self):
        """
        Test that supporting_files is not overwritten when provided initially.
        """
        config = JobbergateConfig(
            supporting_files_output_name=dict(foo="string", bar="string"), supporting_files=["baz"]
        )

        assert config.supporting_files == ["baz"]

    def test_get_supporting_files_from_supporting_files_output_name__missing_value(self):
        """
        Test that supporting_files is still None if supporting_files_output_name is not provided.
        """
        config = JobbergateConfig()

        assert config.supporting_files is None

    def test_supporting_files_output_name_provided_as_string(self):
        """
        Test that string values at supporting_files_output_name are converted to lists of strings.
        """
        config = JobbergateConfig(supporting_files_output_name=dict(foo="bar", baz=["qux"]))

        assert config.supporting_files_output_name == dict(foo=["bar"], baz=["qux"])

    def test_remove_leading_template_on_path__default_template(self):
        """
        Test that the leading "templates/" is removed from the field default_template.
        """
        config = JobbergateConfig(default_template="templates/test_job_script.sh")

        assert config.default_template == "test_job_script.sh"

    def test_remove_leading_template_on_path__supporting_files_output_name(self):
        """
        Test that the leading "templates/" is removed from each key on the dict supporting_files_output_name.
        """
        config = JobbergateConfig(
            supporting_files_output_name={"templates/test_job_script.sh": ["string"]},
        )

        assert config.supporting_files_output_name == {"test_job_script.sh": ["string"]}

    def test_remove_leading_template_on_path__supporting_files(self):
        """
        Test that the leading "templates/" is removed from each entry on the list supporting_files.
        """
        config = JobbergateConfig(supporting_files=["templates/test_job_script.sh"])

        assert config.supporting_files == ["test_job_script.sh"]

    def test_remove_leading_template_on_path__template_files(self):
        """
        Test that the leading "templates/" is removed from each entry on the list template_files.
        """
        config = JobbergateConfig(template_files=["templates/test_job_script.sh"])

        assert config.template_files == ["test_job_script.sh"]


def test_application_config(reference_application_config, dummy_application_config):
    """
    Test that the business logic to fill missing values in the application config is working.
    """
    desired_config = ApplicationConfig.get_from_yaml_file(reference_application_config)
    actual_config = ApplicationConfig.get_from_yaml_file(dummy_application_config)

    assert actual_config == desired_config



"""
Provide a convenience class for managing job-script files.
"""

from pathlib import Path
from typing import Any, Dict, List, Optional, cast

from buzz import Buzz, require_condition
from file_storehouse import FileManager
from jinja2 import Template
from loguru import logger
from pydantic import BaseModel, root_validator

from jobbergate_api.apps.applications.application_files import ApplicationFiles
from jobbergate_api.apps.applications.schemas import ApplicationConfig
from jobbergate_api.config import settings
from jobbergate_api.s3_manager import IO_TRANSFORMATIONS, file_manager_factory, s3_client

JOBSCRIPTS_WORK_DIR = "job-scripts"
JOBSCRIPTS_MAIN_FILE_FOLDER = "main-file"
JOBSCRIPTS_SUPPORTING_FILES_FOLDER = "supporting-files"


class JobScriptCreationError(Buzz):
    """Raise exception when facing any error creating a job-script from application files."""


def flatten_param_dict(param_dict: Dict[str, Any]) -> Dict[str, Any]:
    """
    Flatten an input dictionary to support the rendering process.

    See the example:

    >>> param_dict = {
    ...     "application_config": {"job_name": "rats", "partitions": [...]},
    ...     "jobbergate_config": {
    ...         "default_template": "test_job_script.sh",
    ...         "supporting_files": [...],
    ...         "supporting_files_output_name": {...},
    ...         "template_files": [...],
    ...         "job_script_name": None,
    ...         "output_directory": ".",
    ...         "partition": "debug",
    ...         "job_name": "rats",
    ...     },
    ... }
    >>> flat_param_dict = flatten_param_dict(param_dict)
    >>> print(flat_param_dict)
    {
        "job_name": "rats",
        "partitions": ["debug", "partition1"],
        "default_template": "test_job_script.sh",
        "supporting_files": ["test_job_script.sh"],
        "supporting_files_output_name": {"test_job_script.sh": [...]},
        "template_files": ["templates/test_job_script.sh"],
        "job_script_name": None,
        "output_directory": ".",
        "partition": "debug",
    }
    """
    param_dict_flat = {}
    for (key, value) in param_dict.items():
        if isinstance(value, dict):
            for nest_key, nest_value in value.items():
                param_dict_flat[nest_key] = nest_value
        else:
            param_dict_flat[key] = value
    return param_dict_flat


def inject_sbatch_params(job_script_data_as_string: str, sbatch_params: List[str]) -> str:
    """
    Inject sbatch params into job script.

    Given the job script as job_script_data_as_string, inject the sbatch params in the correct location.
    """
    logger.debug("Preparing to inject sbatch params into job script")

    if not sbatch_params:
        logger.warning("Sbatch param list is empty")
        return job_script_data_as_string

    first_sbatch_index = job_script_data_as_string.find("#SBATCH")
    string_slice = job_script_data_as_string[first_sbatch_index:]
    line_end = string_slice.find("\n") + first_sbatch_index + 1

    inner_string = ""
    for parameter in sbatch_params:
        inner_string += "#SBATCH " + parameter + "\n"

    new_job_script_data_as_string = (
        job_script_data_as_string[:line_end] + inner_string + job_script_data_as_string[line_end:]
    )

    logger.debug("Done injecting sbatch params into job script")
    return new_job_script_data_as_string


class JobScriptFiles(BaseModel):
    """
    Model containing job-script files.

    The folder structure aims to mimic the one specified by the application code.
    In this way, the ``output_directory`` and file names for the main file
    and supporting files are application specific.

    To differentiate the files, they are stored in two folders, ``main-file`` and
    ``supporting-files`` and the paths used in this model are relative the those
    folders. See the example:

        job-scripts
        └───<id>
            └───main-file
            │   └───<output-directory>
            │       └───<main-file-name>
            │
            └───supporting-files
                └───<output-directory>
                    │   <supporting-file-name-1>
                    │   <supporting-file-name-2>
                    │   <supporting-file-name-3>
                    └───...

    In this way, services that consume this information can maintain the
    structure as well (like the user interface and the agents).
    """

    main_file_path: Path
    files: Dict[Path, str]

    @root_validator(pre=False, skip_on_failure=True)
    def check_main_file_path_is_in_files_keys(cls, values):
        """
        Validate the model.

        main_file_path should be found among the files.
        """
        if values["main_file_path"] not in values["files"].keys():
            raise ValueError("main_file_path is not a valid key on the dict files")
        return values

    @property
    def main_file(self):
        """
        Obtain the main file with this helper property.
        """
        return self.files.get(self.main_file_path)

    @classmethod
    def get_from_s3(cls, job_script_id: int):
        """
        Alternative method to initialize the model getting the objects from S3.
        """
        logger.debug(f"Getting job-script files from S3: {job_script_id=}")
        file_manager = cls.file_manager_factory(job_script_id)

        files = {}
        main_file_path = None
        main_file_counter = 0

        for s3_path in file_manager.keys():
            foldername = s3_path.parts[0]
            dict_path = s3_path.relative_to(foldername)
            if foldername == JOBSCRIPTS_MAIN_FILE_FOLDER:
                files[dict_path] = file_manager.get(s3_path)
                main_file_counter += 1
                main_file_path = dict_path
            elif foldername == JOBSCRIPTS_SUPPORTING_FILES_FOLDER:
                files[dict_path] = file_manager.get(s3_path)

        require_condition(
            main_file_counter == 1,
            f"One and only one main file is expected for a job-script, found {main_file_counter}",
            ValueError,
        )

        jobscript_files = cls(main_file_path=main_file_path, files=files)

        logger.debug(f"Done getting job-script files from S3: {job_script_id=}")

        return jobscript_files

    @classmethod
    def delete_from_s3(cls, job_script_id: int):
        """
        Delete the files associated with the given id.
        """
        file_manager = cls.file_manager_factory(job_script_id)
        logger.debug(
            f"Deleting from S3 the files associated to {job_script_id=}. "
            f"Files to be deleted: {', '.join(map(str, file_manager.keys()))}"
        )
        file_manager.clear()
        logger.debug(f"Files were deleted for {job_script_id=}")

    def write_to_s3(self, job_script_id: int, *, remove_previous_files: bool = True):
        """
        Write to s3 the files associated with a given id.
        """
        logger.debug(f"Writing job-script files to S3: {job_script_id=}")

        file_manager = self.file_manager_factory(job_script_id)

        if remove_previous_files and file_manager:
            self.delete_from_s3(job_script_id)

        for dict_path, content in self.files.items():
            if dict_path == self.main_file_path:
                s3_path = JOBSCRIPTS_MAIN_FILE_FOLDER / dict_path
            else:
                s3_path = JOBSCRIPTS_SUPPORTING_FILES_FOLDER / dict_path
            file_manager[s3_path] = content

        logger.debug(f"Done writing job-script files to S3 for {job_script_id=}")

    @classmethod
    def render_from_application(
        cls,
        application_files: ApplicationFiles,
        user_supplied_parameters: Dict[str, Any] = None,
        sbatch_params: List[str] = None,
    ):
        """
        Render JobScriptFiles from ApplicationFiles.
        """
        logger.debug("Rendering job-script files from an application")

        with JobScriptCreationError.check_expressions(
            main_message="One or more application files are missing",
        ) as check:
            check(application_files.config_file, "Application config file was not found")
            check(application_files.source_file, "Application source file was not found")
            check(application_files.templates, "No template file was found")

        with JobScriptCreationError.handle_errors(
            "Error while parsing the config-file and/or job-script params",
        ):
            app_config = ApplicationConfig.get_from_yaml_file(
                application_files.config_file, user_supplied_parameters  # type: ignore
            )

        default_template_name: str = JobScriptCreationError.enforce_defined(
            app_config.jobbergate_config.default_template,
        )

        JobScriptCreationError.require_condition(
            default_template_name in application_files.templates,
            "Selected template {selected} not found in available templates: {templates}".format(
                selected=default_template_name,
                templates=", ".join(application_files.templates.keys()),
            ),
        )

        if app_config.jobbergate_config.supporting_files_output_name:
            with JobScriptCreationError.check_expressions(
                main_message="One or more supporting files are missing"
            ) as check:
                for supporting_file in app_config.jobbergate_config.supporting_files_output_name.keys():
                    check(supporting_file in application_files.templates, f"{supporting_file=} was not found")

        with JobScriptCreationError.handle_errors("Error rendering the main file"):

            param_dict_flat = flatten_param_dict(app_config.dict())

            output_directory: Optional[str] = app_config.jobbergate_config.output_directory

            # The output_directory can be None here if the Application *explicitly* sets it to None.
            if not output_directory:
                output_directory = "."

            main_file_path = Path(
                output_directory,
                default_template_name.rstrip(".j2").rstrip(".jinja2"),
            )

            main_file_content = Template(
                application_files.templates[default_template_name],
            ).render(data=param_dict_flat)

        with JobScriptCreationError.handle_errors("Error while injecting the sbatch params"):
            if sbatch_params:
                main_file_content = inject_sbatch_params(main_file_content, sbatch_params)

        with JobScriptCreationError.handle_errors("Error while creating JobScriptFiles object"):
            jobscript_files = cls(
                main_file_path=main_file_path,
                files={main_file_path: main_file_content},
            )

        with JobScriptCreationError.handle_errors("Error while rendering the supporting files"):
            output_name_mapping = app_config.jobbergate_config.supporting_files_output_name
            if output_name_mapping:
                for template_name, supporting_filename_list in output_name_mapping.items():

                    for supporting_filename in supporting_filename_list:
                        path = Path(output_directory, supporting_filename)

                        jobscript_files.files[path] = Template(
                            application_files.templates[template_name],
                        ).render(data=param_dict_flat)

        logger.debug("Done rendering job-script files from an application")

        return jobscript_files

    @classmethod
    def file_manager_factory(cls, job_script_id: int) -> FileManager:
        """
        Build an application file manager.
        """
        return cast(
            FileManager,
            file_manager_factory(
                id=job_script_id,
                s3_client=s3_client,
                bucket_name=settings.S3_BUCKET_NAME,
                work_directory=Path(JOBSCRIPTS_WORK_DIR),
                manager_cls=FileManager,
                transformations=IO_TRANSFORMATIONS,
            ),
        )



"""
Describe constants for the job_submissions module.
"""


from enum import Enum


class JobSubmissionStatus(str, Enum):
    """
    Defines the set of possible statuses for a Job Submission.
    """

    CREATED = "CREATED"
    SUBMITTED = "SUBMITTED"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    UNKNOWN = "UNKNOWN"
    REJECTED = "REJECTED"

    @classmethod
    def pretty_list(cls):
        """
        Return a comma-separated list of possible statuses.
        """
        return ", ".join(str(e) for e in cls)



"""
Provide unit tests for the main app.
"""
import pytest
from fastapi import status
from httpx import AsyncClient


@pytest.mark.asyncio
async def test_health_check(client: AsyncClient):
    """
    Test the health check route.

    This test ensures the API has a health check path configured properly, so
    the production and staging environments can configure the load balancing
    """
    response = await client.get("/jobbergate/health")

    assert response.status_code == status.HTTP_204_NO_CONTENT



"""
Test application files.
"""

import contextlib

import pytest
from fastapi import UploadFile

from jobbergate_api.apps.applications.application_files import ApplicationFiles
from jobbergate_api.apps.applications.constants import (
    APPLICATION_CONFIG_FILE_NAME,
    APPLICATION_SOURCE_FILE_NAME,
    APPLICATION_TEMPLATE_FOLDER,
    APPLICATIONS_WORK_DIR,
)
from jobbergate_api.apps.applications.schemas import ApplicationConfig, JobbergateConfig


class TestApplicationConfig:
    """
    Test the application config base model.
    """

    def test_get_from_yaml_file_without_extra_params__success(self, dummy_application_config):
        """
        Test the model can be loaded from a yaml file without extra params.
        """
        desired_application_config = ApplicationConfig(
            application_config={"job_name": "rats", "partitions": ["debug", "partition1"]},
            jobbergate_config=JobbergateConfig(
                default_template="test_job_script.sh",
                supporting_files=["test_job_script.sh"],
                supporting_files_output_name={"test_job_script.sh": ["support_file_b.py"]},
                template_files=["templates/test_job_script.sh"],
                job_script_name=None,
                output_directory=".",
            ),
        )

        actual_application_config = ApplicationConfig.get_from_yaml_file(dummy_application_config)

        assert actual_application_config == desired_application_config

    def test_get_from_yaml_file_including_extra_params__success(self, dummy_application_config):
        """
        Test the model can be loaded from a yaml file with extra params.
        """
        user_supplied_parameters = dict(
            application_config=dict(job_name="testing"),
            jobbergate_config=dict(default_template="another_test_job_script.sh"),
        )

        desired_application_config = ApplicationConfig(
            application_config={"job_name": "testing"},
            jobbergate_config=JobbergateConfig(
                default_template="another_test_job_script.sh",
            ),
        )

        actual_application_config = ApplicationConfig.get_from_yaml_file(
            dummy_application_config, user_supplied_parameters
        )

        assert actual_application_config == desired_application_config

    def test_get_from_yaml_file_validation_error(self):
        """
        Test that ValueError is raised when the file does not have the necessary schema.
        """
        with pytest.raises(ValueError):
            ApplicationConfig.get_from_yaml_file("key: value")


@pytest.fixture
def make_uploaded_files():
    """
    Provide a fixture to use as a context manager that builds a list of uploaded files.
    """

    @contextlib.contextmanager
    def _helper(*file_paths):
        """
        Context manager that opens the file(s) and yields the ``files`` param from it.
        """
        with contextlib.ExitStack() as stack:
            yield [
                UploadFile(path.name, stack.enter_context(open(path, "rb")), "text/plain")
                for path in file_paths
            ]

    return _helper


@pytest.fixture(scope="function")
def make_uploaded_files_filled(
    make_dummy_file,
    make_uploaded_files,
    dummy_application_source_file,
    dummy_template,
    dummy_application_config,
):
    """
    Fixture to make a list of uploaded files.
    """
    with make_uploaded_files(
        make_dummy_file("jobbergate.py", content=dummy_application_source_file),
        make_dummy_file("template-1.j2", content=dummy_template),
        make_dummy_file("template-2.jinja2", content=dummy_template),
        make_dummy_file("jobbergate.yaml", content=dummy_application_config),
    ) as uploaded_files:
        yield uploaded_files


class TestApplicationFiles:
    """
    Test application files.
    """

    def test_get_application_files_from_upload_files(
        self,
        make_uploaded_files_filled,
        dummy_application_source_file,
        dummy_template,
        dummy_application_config,
    ):
        """
        Test that ApplicationFiles can be obtained from a list of uploaded files.
        """
        application_files = ApplicationFiles.get_from_upload_files(make_uploaded_files_filled)

        assert isinstance(application_files, ApplicationFiles)

        assert application_files.config_file == dummy_application_config
        assert application_files.source_file == dummy_application_source_file
        assert application_files.templates == {
            "template-1.j2": dummy_template,
            "template-2.jinja2": dummy_template,
        }

    def test_write_application_files_to_s3(
        self,
        make_uploaded_files_filled,
        dummy_application_source_file,
        dummy_template,
        dummy_application_config,
        mocked_file_manager_factory,
    ):
        """
        Test if the applications files are written to S3 as expected.

        To do so, a list of uploaded files is created, than  we verify if each file
        went to the correct file manager.
        """
        application_id = 1

        ApplicationFiles.get_from_upload_files(
            make_uploaded_files_filled,
        ).write_to_s3(application_id)

        work_dir = mocked_file_manager_factory / APPLICATIONS_WORK_DIR / str(application_id)
        templates_dir = work_dir / APPLICATION_TEMPLATE_FOLDER

        application_source_path = work_dir / APPLICATION_SOURCE_FILE_NAME
        assert application_source_path.is_file()
        assert application_source_path.read_text() == dummy_application_source_file

        application_config_path = work_dir / APPLICATION_CONFIG_FILE_NAME
        assert application_config_path.is_file()
        assert application_config_path.read_text() == dummy_application_config

        assert templates_dir.is_dir()

        template_path_1 = templates_dir / "template-1.j2"
        assert template_path_1.is_file()
        assert template_path_1.read_text() == dummy_template

        template_path_2 = templates_dir / "template-2.jinja2"
        assert template_path_2.is_file()
        assert template_path_2.read_text() == dummy_template

    def test_get_application_files_from_s3(
        self,
        dummy_application_source_file,
        dummy_application_config,
        dummy_template,
        mocked_file_manager_factory,
    ):
        """
        Test if the application files are loaded from S3 as expected.

        To do so, dummy test files are added to the file managers and than loaded.
        """
        application_id = 1
        work_dir = mocked_file_manager_factory / APPLICATIONS_WORK_DIR / str(application_id)
        templates_dir = work_dir / APPLICATION_TEMPLATE_FOLDER
        templates_dir.mkdir(parents=True, exist_ok=True)

        application_path = work_dir / APPLICATION_SOURCE_FILE_NAME
        application_path.write_text(dummy_application_source_file)

        config_path = work_dir / APPLICATION_CONFIG_FILE_NAME
        config_path.write_text(dummy_application_config)

        template_path_1 = templates_dir / "template-1.j2"
        template_path_1.write_text(dummy_template)

        template_path_2 = templates_dir / "template-2.jinja2"
        template_path_2.write_text(dummy_template)

        application_files = ApplicationFiles.get_from_s3(application_id)

        assert isinstance(application_files, ApplicationFiles)

        assert application_files.source_file == dummy_application_source_file

        assert application_files.config_file == dummy_application_config

        assert application_files.templates == {
            "template-1.j2": dummy_template,
            "template-2.jinja2": dummy_template,
        }

    def test_delete_application_files_from_s3(
        self,
        dummy_application_source_file,
        dummy_template,
        mocked_file_manager_factory,
    ):
        """
        Test if the applications files are deleted from S3 as expected.

        To do so, test files are added to both file managers. Once one application id
        is deleted, all templated files related to it are deleted as well as its source
        file. Source files from any other id are expected to be preserved.
        """
        application_id = 1

        work_dir = mocked_file_manager_factory / APPLICATIONS_WORK_DIR / str(application_id)
        templates_dir = work_dir / APPLICATION_TEMPLATE_FOLDER
        templates_dir.mkdir(parents=True, exist_ok=True)

        application_path = work_dir / APPLICATION_SOURCE_FILE_NAME
        application_path.write_text(dummy_application_source_file)

        template_path_1 = templates_dir / "template-1.j2"
        template_path_1.write_text(dummy_template)

        template_path_2 = templates_dir / "template-2.jinja2"
        template_path_2.write_text(dummy_template)

        ApplicationFiles.delete_from_s3(application_id)

        assert not application_path.is_file()
        assert not template_path_1.is_file()
        assert not template_path_2.is_file()

    def test_delete_just_act_on_one_id(
        self,
        make_uploaded_files_filled,
        mocked_file_manager_factory,
    ):
        """
        Test that delete does not delete any file from any other application.
        """
        remaining_application_id = 1
        another_application_id = 2

        desired_application_files = ApplicationFiles.get_from_upload_files(
            make_uploaded_files_filled,
        )

        desired_application_files.write_to_s3(remaining_application_id)
        desired_application_files.write_to_s3(another_application_id)

        ApplicationFiles.delete_from_s3(another_application_id)

        deleted_application_files = ApplicationFiles.get_from_s3(another_application_id)
        assert deleted_application_files.dict(exclude_unset=True, exclude_defaults=True) == {"templates": {}}

        actual_application_files = ApplicationFiles.get_from_s3(remaining_application_id)
        assert desired_application_files == actual_application_files

    def test_complete_workflow(self, mocked_file_manager_factory, make_uploaded_files_filled):
        """
        End-to-end test, getting application files from uploaded files, than write/read them.
        """
        application_id = 1

        desired_application_files = ApplicationFiles.get_from_upload_files(
            make_uploaded_files_filled,
        )

        desired_application_files.write_to_s3(application_id)

        actual_application_files = ApplicationFiles.get_from_s3(application_id)

        assert desired_application_files == actual_application_files
